#!/bin/sh

echo "/* Automatically generated by generate-configlist.sh */"
echo

print_config_list () {
	cat <<EOF
static const char *config_name_list[] = {
EOF
	grep -h '^[a-zA-Z].*\..*::$' Documentation/*config.txt Documentation/config/*.txt |
	sed '/deprecated/d; s/::$//; s/,  */\n/g' |
	sort |
	sed 's/^.*$/	"&",/'
	cat <<EOF
	NULL,
};
EOF
}

echo
print_config_list
